/***************************************************************************
 *  helicontrol - c/c++ control api for silverlit helicopters/quadcopters  *
 *  Copyright (C) 2008 by                                                  *
 *  Joint Robotics Lab                                                     *
 *  Goethe-University Frankfurt, Germany                                   *
 *  http://www.jrl.cs.uni-frankfurt.de                                     *
 *                                                                         *
 *  This library is free software; you can redistribute it and/or          *
 *  modify it under the terms of the GNU General Public                    *
 *  License as published by the Free Software Foundation                   *
 *  version 2 of the License.                                              *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU      *
 *  General Public License for more details.                               *
 *                                                                         *
 *  You should have received a copy of the GNU General Public              *
 *  License along with this library; if not, write to the Free Software    *
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA02110-1301 USA*
 ***************************************************************************/

/* $Id: helidemo_sdl.c 268 2008-12-17 21:52:22Z holgerf $ */

/** @file
 * This is a simple demo which shows how to work with libhelicontrol.
 * It implements a simple remote control, similar to helictrl.cc in ../tools
 * but with joystick support (SDL, not available on win32/vs2008).
 * @example helidemo_sdl.c
*/




#include <stdio.h>
#include <stdint.h>
#include <getopt.h>
#include <errno.h>
#include <unistd.h>
#include <SDL/SDL.h>
#include <helicontrol.h>


#define bound(min,num,max) ((num) >= (min) ? ((num) < (max) ? (num) : (max)) : (min))

#define JOYSTICK					0
#define JOYSTICK_THROTTLE_MAX		-32768
#define JOYSTICK_YAW_MAX			-32768
#define JOYSTICK_PITCH_MAX			-32768
#define JOYSTICK_ROLL_MAX			32768
#define JOYSTICK_AXIS_SAUCER_ROLL	2
#define JOYSTICK_AXIS_SAUCER_YAW	0
#define JOYSTICK_AXIS_THROTTLE		1
#define JOYSTICK_AXIS_YAW			2
#define JOYSTICK_AXIS_PITCH			3
#define JOYSTICK_BUTTON_TRIM_L		6
#define JOYSTICK_BUTTON_TRIM_R		7
#define JOYSTICK_BUTTON_FIRE1		4
#define JOYSTICK_BUTTON_FIRE2		5
#define JOYSTICK_BUTTON_PITCH1		4
#define JOYSTICK_BUTTON_PITCH2		5

#define INTERVAL					25000

#define MODE_NORMAL					0
#define MODE_JOYSTICK				1
#define MODE_STOP					2

static struct option long_options[] =
{
    {"protocol",	required_argument,	0, 'p'},
    {"channel",		required_argument,	0, 'c'},
    {"throttle",	required_argument,	0, 't'},
    {"trim",		required_argument,	0, 'm'},
    {"yaw",			required_argument,	0, 'y'},
    {"fire",		required_argument,	0, 'f'},
    {"pitch",		required_argument,	0, 'i'},
    {"roll",		required_argument,	0, 'r'},
    {"stop",		no_argument,		0, 's'},
    {"joystick",	no_argument,		0, 'j'},
    {"help",		no_argument,		0, 'h'},
    {0, 0, 0, 0}
};

static void usage(void)
{
    puts("Usage: -p p -c a -t 5 -y 0");
    puts("-p/--protocol:   transmission protocol (p(icooz)/c(hallenger)/u(ranus)/(t)andemz/(s)aucer)");
    puts("-c/--channel:    channel (a/b/c)");
    puts("-t/--throttle:   throttle (0..15)");
    puts("-m/--trim:       trim (l/r)");
    puts("-y/--yaw:        yaw (-3(left)..3(right))");
    puts("-s/--stop:       stop repeating the last signal");
    puts("-j/--joystick:   use joystick/gamepad for interactive input");
    puts("-f/--fire:       fire (0/1, challenger only)");
    puts("-i/--pitch:      pitch control (uranus/tandemz/saucer only)");
    puts("-r/--roll:       roll control (flying saucer only)");
}

struct usb_dev_handle *handle;

static void send_frame(char protocol, char channel, char throttle, char trim, char yaw, char pitch, char special)
{
    int error;

    switch (protocol)
    {
        case PROTO_PICOOZ:
            error = HC_Send_Picooz(handle, bound(0, channel-'a', 2), throttle, yaw, trim);
            break;
        case PROTO_CHALLENGER:
            error = HC_Send_Challenger(handle, bound(1, channel-'a'+1, 2), throttle, yaw, trim, special);
            break;
        case PROTO_URANUS:
            error = HC_Send_Uranus(handle, bound(0, channel-'a', 2), throttle, yaw, pitch, trim, special);
            break;
        case PROTO_TANDEMZ:
            error = HC_Send_Tandemz(handle, bound(0, channel-'a', 2), throttle, yaw, pitch, trim, special);
            break;
        case PROTO_SAUCER:
            error = HC_Send_Saucer(handle, throttle, yaw, pitch, special);
            break;
    }
    if (error < 0)
    {
        puts("USB_control_msg error!");
        exit(EXIT_FAILURE);
    }
}

int main(int argc, char **argv)
{
    char channel, throttle, trim, yaw, pitch, special, protocol = PROTO_CHALLENGER;
    int c, mode = MODE_NORMAL, option_index = 0;
    SDL_Joystick *joy;

    if (argc == 1)
    {
        usage();
        exit(EXIT_FAILURE);
    }

    channel = throttle = trim = yaw = pitch = special = 0;

    while (1)
    {
        c = getopt_long(argc, argv, "p:c:t:m:y:r:f:i:sjh", long_options, &option_index);
        if (c == -1) break;
        switch (c)
        {
            case 'h':
                usage();
                exit(EXIT_FAILURE);
            case 'p':
                switch (optarg[0])
                {
                    case 'p':
                        protocol = PROTO_PICOOZ;
                        break;
                    case 'c':
                        protocol = PROTO_CHALLENGER;
                        break;
                    case 'u':
                    case 'a':
                        protocol = PROTO_URANUS;
                        break;
                    case 't':
                        protocol = PROTO_TANDEMZ;
                        break;
                    case 's':
                        protocol = PROTO_SAUCER;
                        break;
                    default:
                        protocol = PROTO_PICOOZ;
                }
                break;
            case 'c':
                channel = optarg[0];
                break;
            case 't':
                throttle = atoi(optarg);
                break;
            case 'm':
                switch (optarg[0])
                {
                    case 'l':
                        trim = -1;
                        break;
                    case 'r':
                        trim = 1;
                        break;
                }
                break;
            case 'y':
                yaw = atoi(optarg);
                break;
            case 'r':
                special = atoi(optarg);
                break;
            case 'i':
                pitch = atoi(optarg);
                break;
            case 'f':
                special = (optarg[0] == '1')?3:0;
                break;
            case 'j':
                mode = MODE_JOYSTICK;
                break;
            case 's':
                mode = MODE_STOP;
                break;
            default:
                abort();
        }
    }
    if (optind < argc)
    {
        printf("Unsupported arguments: ");
        while (optind < argc)
            printf("%s ", argv[optind++]);
        putchar('\n');
        exit(EXIT_FAILURE);
    }

    handle = HC_Init();
    if (handle == NULL)
    {
        puts("USB device handle not found!");
        exit(EXIT_FAILURE);
    }

    if (mode == MODE_JOYSTICK) /* joystick */
    {
        SDL_Init(SDL_INIT_JOYSTICK);
        joy = SDL_JoystickOpen(JOYSTICK);
        if (!joy)
        {
            puts("Could not open joystick!");
            usb_close(handle);
            exit(EXIT_FAILURE);
        }

        if (HC_CMD_Repeat(handle, FALSE) < 0)
            puts("USB_control_msg error!");

        while (1)
        {
            SDL_JoystickUpdate();

            throttle = (int)(SDL_JoystickGetAxis(joy, JOYSTICK_AXIS_THROTTLE) * 16.0/JOYSTICK_THROTTLE_MAX);
            yaw = (int)(SDL_JoystickGetAxis(joy, JOYSTICK_AXIS_YAW) * 4.0/JOYSTICK_YAW_MAX);

            if (SDL_JoystickGetButton(joy, JOYSTICK_BUTTON_FIRE1) || SDL_JoystickGetButton(joy, JOYSTICK_BUTTON_FIRE2))
                special = 3;

            if (SDL_JoystickGetButton(joy, JOYSTICK_BUTTON_TRIM_L))
                trim = -1;
            else if (SDL_JoystickGetButton(joy, JOYSTICK_BUTTON_TRIM_R))
                trim = 1;


            if (protocol == PROTO_URANUS)
            {
                if (SDL_JoystickGetButton(joy, JOYSTICK_BUTTON_PITCH1))
                    pitch = 1;
                else if (SDL_JoystickGetButton(joy, JOYSTICK_BUTTON_PITCH2))
                    pitch ^= 3;
            }
            else if (protocol == PROTO_TANDEMZ)
            {
                pitch = 2;
                trim = (int)(SDL_JoystickGetAxis(joy, JOYSTICK_AXIS_PITCH) * 16.5/JOYSTICK_PITCH_MAX + 15.5);
                yaw = (int)(SDL_JoystickGetAxis(joy, JOYSTICK_AXIS_YAW) * -16.5/JOYSTICK_YAW_MAX + 15.5);
            }
            else if (protocol == PROTO_SAUCER)
            {
                throttle = (int)(SDL_JoystickGetAxis(joy, JOYSTICK_AXIS_THROTTLE) * 15.0/JOYSTICK_THROTTLE_MAX);
                pitch = (int)(SDL_JoystickGetAxis(joy, JOYSTICK_AXIS_PITCH) * -7.5/JOYSTICK_PITCH_MAX);
                yaw = (int)(SDL_JoystickGetAxis(joy, JOYSTICK_AXIS_SAUCER_YAW) * -7.5/JOYSTICK_YAW_MAX);
                special = (int)(SDL_JoystickGetAxis(joy, JOYSTICK_AXIS_SAUCER_ROLL) * 7.5/JOYSTICK_ROLL_MAX); /* roll */
            }

            if (SDL_JoystickGetButton(joy, JOYSTICK_BUTTON_FIRE1) || SDL_JoystickGetButton(joy, JOYSTICK_BUTTON_FIRE2))
            {
                if (HC_CMD_Stop(handle) < 0)
                    puts("USB_control_msg error!");
            }
            else
                send_frame(protocol, channel, throttle, trim, yaw, pitch, special);

            usleep(INTERVAL);
        }
    }
    else if (mode == MODE_STOP) /* stop */
    {
        if (HC_CMD_Stop(handle) < 0)
            puts("USB_control_msg error!");
    }
    else
    {
        if (HC_CMD_Repeat(handle, TRUE) < 0)
            puts("USB_control_msg error!");
        send_frame(protocol, channel, throttle, trim, yaw, pitch, special);
    }

    HC_Close(handle);

    exit(EXIT_SUCCESS);
}
/* EOF */
