/* $Id: helicontrol_firmware.c 267 2008-12-17 19:47:37Z siegl $ */

// helicontrol.c
// remote control for Silverlit Picooz, Challenger, Atlas/Uranus, TandemZ
// & Alien Air Flying Saucer

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <avr/eeprom.h>
#include <stdlib.h>
#include "defines.h"
#include <util/delay.h>
#include "helicontrol_firmware.h"
#include "frame.h"
#include "usart_drv.h"
#include "usb_drv.h"
#include "macros.h"

#define min(A, B) (((A)<(B))?(A):(B))
#define max(A, B) (((A)>(B))?(A):(B))

// IR-LED control
#define ir_on(A)		SetBit(DDRB, PINB4); _delay_ms((A))
#define ir_off(A)		ClearBit(DDRB, PINB4); _delay_ms((A))

// status LED control (AVR is used as sink, so LED is on when PORTE:7 is 0)
#define led_on()		ClearBit(PORTE, PINE7)
#define led_off()		SetBit(PORTE, PINE7)

// wakeup timer
#define timer_start()	TCNT1 = 0; SetBit(TCCR1B, CS12)
#define timer_stop()	ClearBit(TCCR1B, CS12); TCNT1 = 0
#define timer_set(A)	OCR1A = (uint16_t)((A)/(1.0/F_CPU*256*1000))

uint8_t protocol;
uint8_t send = FALSE;
uint8_t autorepeat = TRUE;
uint16_t interval = DEFAULT_INTERVAL_MS;
uint8_t framelength;
double pause_roll, pause_pitch, pause_throttle, pause_yaw;	// Flying Saucer

void send_header(void)
{
    switch (protocol)
    {
        case PROTO_PICOOZ:
            ir_on(1.9);
            ir_off(0.65);
            break;
        case PROTO_CHALLENGER:
        case PROTO_URANUS:
            ir_on(0.6);
            ir_off(0.6);
            ir_on(1.28);
            ir_off(3);
            ir_on(1.92);
            ir_off(0.6);
            break;
        case PROTO_TANDEMZ:
            ir_on(0.68);
            ir_off(1.2);
            ir_on(2);
            ir_off(1.28);
            break;
        case PROTO_IBIRD:
            ir_on(2);
            ir_off(2);
            ir_on(2);
            ir_off(1);
            break;
        case PROTO_SAUCER:
            return;
    }
}

void send_id(void)
{
    if (protocol != PROTO_PICOOZ)
        return;

    ir_on(0.65);
    ir_off(0.65);
    ir_on(0.65);
    ir_off(0.65);
    ir_on(1.15);
    ir_off(0.65);
    ir_on(1.15);
    ir_off(0.65);
}

void send_stopbit(void)
{
    switch (protocol)
    {
        case PROTO_PICOOZ:
            ir_on(0.65);
            break;
        case PROTO_CHALLENGER:
        case PROTO_URANUS:
            ir_on(0.61);
            break;
        case PROTO_TANDEMZ:
            ir_on(0.65);
            break;
        case PROTO_IBIRD:
            ir_on(0.3);
            break;
        case PROTO_SAUCER:
            return;
    }
    // wait a little longer than the protocol requires after switching the LEDs off,
    // just in case another control msg is already waiting
    ir_off(3);
}

void send_data(void)
{
    uint32_t i;

    switch (protocol)
    {
        case PROTO_PICOOZ:
            for (i=1UL<<(SIZE_PICOOZ-1); i; i>>=1)
                if (frame.raw&i)
                {
                    ir_on(1.14);
                    ir_off(1.14);
                }
                else
                {
                    ir_on(0.57);
                    ir_off(0.57);
                }
            break;
        case PROTO_CHALLENGER:
        case PROTO_URANUS:
            for (i=1UL<<(SIZE_CHALLENGER-1); i; i>>=1)
                if (frame.raw&i)
                {
                    ir_on(1.22);
                    ir_off(1.22);
                }
                else
                {
                    ir_on(0.61);
                    ir_off(0.61);
                }
            break;
        case PROTO_TANDEMZ:
            for (i=1UL<<(SIZE_TANDEMZ-1); i; i>>=1)
                if (frame.raw&i)
                {
                    ir_on(1.2);
                    ir_off(0.6);
                }
                else
                {
                    ir_on(0.65);
                    ir_off(0.65);
                }
            break;
        case PROTO_IBIRD:
            for (i=1UL<<(SIZE_IBIRD-1); i; i>>=1)
                if (frame.raw&i)
                {
                    ir_on(1.4);
                    ir_off(1.4);
                }
                else
                {
                    ir_on(0.68);
                    ir_off(0.8);
                }
            break;
        case PROTO_SAUCER:
#define PULSE_LENGTH 0.24
            ir_on(PULSE_LENGTH);
            ir_off(pause_roll);
            ir_on(PULSE_LENGTH);
            ir_off(pause_pitch);
            ir_on(PULSE_LENGTH);
            ir_off(pause_throttle);
            ir_on(PULSE_LENGTH);
            ir_off(pause_yaw);
            ir_on(PULSE_LENGTH);
            ir_off(0.1);		// switch LEDs off, better safe than sorry...
            break;
    }
}

uint8_t checksum(void)
{
    uint32_t data, i;
    uint8_t pos=0, cs=0;

    data = frame.raw;

    switch (protocol)
    {
        case PROTO_PICOOZ:
            data >>= 2;
            for (i=1UL<<(SIZE_PICOOZ-3); i; i>>=1, ++pos)
                if (data&i)
                    cs += (pos&1)?3:2;
            cs = (cs&2)>>1 | (cs&1)<<1;
            break;
        case PROTO_CHALLENGER:
        case PROTO_URANUS:
            data >>= 2;
            for (i=1UL<<(SIZE_CHALLENGER-3); i; i>>=1, ++pos)
                if (data&i)
                    cs ^= (pos&1)?3:1;
            break;
        case PROTO_TANDEMZ:
            cs = frame.tandemz.channel
                 + frame.tandemz.light*4
                 + (frame.tandemz.throttle&8)/2
                 + (frame.tandemz.throttle&7)
                 + frame.tandemz.pitch
                 + (frame.tandemz.ptrim>>3)
                 + (frame.tandemz.ptrim&7)
                 + (frame.tandemz.yaw>>3)
                 + (frame.tandemz.yaw&7);
            cs &= 7;
            break;
        case PROTO_IBIRD:
            break;
    }

    return cs;
}


void ProcessUserCommand(uint8_t request, uint16_t value, uint16_t index)
{
    switch (request)
    {
        case PROTO_PICOOZ:
            protocol = PROTO_PICOOZ;
            framelength = SIZE_PICOOZ;
            frame.usb.value = value;
            frame.usb.index = index;
            frame.picooz.throttle = min(14, frame.picooz.throttle);
            frame.picooz.yaw = -frame.picooz.yaw;
            frame.picooz.checksum = checksum();
            timer_set(interval);
            send = TRUE;
            break;
        case PROTO_CHALLENGER:
            protocol = PROTO_CHALLENGER;
            framelength = SIZE_CHALLENGER;
            frame.usb.value = value;
            frame.usb.index = index;
            frame.challenger.checksum = checksum();
            timer_set(interval);
            send = TRUE;
            break;
        case PROTO_URANUS:
            protocol = PROTO_URANUS;
            framelength = SIZE_URANUS;
            frame.usb.value = value;
            frame.usb.index = index;
            frame.uranus.checksum = checksum();
            timer_set(interval);
            send = TRUE;
            break;
        case PROTO_TANDEMZ:
            protocol = PROTO_TANDEMZ;
            framelength = SIZE_TANDEMZ;
            frame.usb.value = value;
            frame.usb.index = index;
            frame.tandemz.ptrim = min(30, frame.tandemz.ptrim);
            frame.tandemz.yaw   = min(30, frame.tandemz.yaw);
            if (frame.tandemz.yaw == 3)
                frame.tandemz.yaw = 2;
            else if (frame.tandemz.yaw == 4 || frame.tandemz.yaw == 5)
                frame.tandemz.yaw = 6;
            else if (frame.tandemz.yaw == 19 || frame.tandemz.yaw == 20)
                frame.tandemz.yaw = 18;
            else if (frame.tandemz.yaw == 21 || frame.tandemz.yaw == 22)
                frame.tandemz.yaw = 23;
            else if (frame.tandemz.yaw == 27)
                frame.tandemz.yaw = 26;
            else if (frame.tandemz.yaw > 27)
                frame.tandemz.yaw = 30;
            frame.tandemz.checksum = checksum();
            timer_set(interval);
            send = TRUE;
            break;
        case PROTO_IBIRD:
            protocol = PROTO_IBIRD;
            framelength = SIZE_IBIRD;
            frame.usb.value = value;
            frame.usb.index = index;
            // frame.ibird.checksum = checksum();
            timer_set(interval);
            send = TRUE;
            break;
        case PROTO_SAUCER:
            protocol = PROTO_SAUCER;
            frame.usb.value = value;
            frame.usb.index = index;

            pause_roll = 1.2 + (max(-7, frame.saucer.roll)+7)/14.0;
            pause_pitch = 1.2 + (max(-7, frame.saucer.pitch)+7)/14.0;
            pause_throttle = 1.7 - frame.saucer.throttle/15.0;
            pause_yaw = 1.2 + (max(-7, frame.saucer.yaw)+7)/14.0;

            // the Flying Saucer faster intervals than Silverlit products, if the interval
            // is too high, the Saucer won't accept any signals (approx. 60ms seems to be
            // the maximum interval allowed)
            timer_set(13);  // ~13ms, like the original remote
            send = TRUE;
            break;
        case CMD_AUTOREPEAT:
            autorepeat = !!value;
            break;
        case CMD_REPEAT_INTERVAL:
            if (value)
                interval = value;
            else
                interval = DEFAULT_INTERVAL_MS;
            if (protocol != PROTO_SAUCER)
                timer_set(interval);
            break;
        case CMD_STOP:
            timer_stop();
            send = FALSE;
            break;
    }
}

// generates a serial number and stores it in eeprom
void set_serial(void)
{
    uint8_t i;
    uint32_t serial;

    // generate new serial, if first 4 bytes of eeprom are blank
    eeprom_read_block(&serial, 0x00, 4);
    if (serial == 0xFFFFFFFF)
    {
        // abuse unused ADC as entropy source
        ADMUX  = 1<<REFS1 | 1<<REFS0;
        ADCSRA = 1<<ADPS2 | 1<<ADEN | 1<<ADSC;
        while (ADCSRA & 1<<ADSC);
        srand(ADCW);

        // disable ADC, as we don't need it any more
        ADMUX = ADCSRA = ADCW = 0;

        for (i=0; i<4; ++i)
            ((char*) &serial)[i] = rand();
        eeprom_write_block(&serial, 0x00, 4);  // store serial
    }
}

void init(void)
{
#ifdef OVERRIDE_CLKDIV8
    // run cpu at full speed (overrides CLKDIV8 fuse)
    CLKPR = 1<<CLKPCE;
    CLKPR = 0;
#endif

    // initialize status LED on PINE7
    DDRE |= (1<<PINE7);
    led_on();

    set_serial();

    // set timer2 as frequency generator to 38kHz on PINB4
    TCCR2A = (1<<WGM21) | (1<<COM2A0);        // CTC mode (see datasheet, section 15.6.2)
    TCCR2B = (1<<CS20);                       // prescaler == 1
    OCR2A = (uint8_t)(F_CPU/(2*1*38000UL)-1); // calculate correct frequency, depending on F_CPU
    DDRB = 0;                                 // deactivate PINB4 by setting it as input

    // prepare timer1 for waking up from sleep mode
    TCCR1B = (1<<WGM12);
    TIMSK1 = (1<<OCIE1A);
    timer_set(interval);

    // disable pull-up resistor
    MCUCR |= (1<<PUD);
    set_sleep_mode(SLEEP_MODE_IDLE);
}

// wakes cpu from sleep mode
ISR(TIMER1_COMPA_vect)
{
    send = TRUE;
}

int main(void)
{
    cli();
    init();
    USART_Init();
    USART_WriteString("\r\n----------------\r\nHeliControl Init\r\n----------------\r\n");
    UsbDevLaunchDevice(false);
    UsbDevWaitStartupFinished(); // sei();

    led_off();

    while (1)
    {
        if (send)
        {
            //cli();
            led_on();
            timer_stop();

            send_header();
            send_id();
            send_data();
            send_stopbit();

            if (autorepeat)
                timer_start();
            send = FALSE;
            led_off();
            //sei();
        }
        else
        {
            sleep_mode();
        }
    }
    return 0;
}
