/***************************************************************************
 *  helicontrol - c/c++ control api for silverlit helicopters/quadcopters  *
 *  Copyright (C) 2008 by                                                  *
 *  Joint Robotics Lab                                                     *
 *  Goethe-University Frankfurt, Germany                                   *
 *  http://www.jrl.cs.uni-frankfurt.de                                     *
 *                                                                         *
 *  This library is free software; you can redistribute it and/or          *
 *  modify it under the terms of the GNU General Public                    *
 *  License as published by the Free Software Foundation                   *
 *  version 2 of the License.                                              *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU      *
 *  General Public License for more details.                               *
 *                                                                         *
 *  You should have received a copy of the GNU General Public              *
 *  License along with this library; if not, write to the Free Software    *
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA02110-1301 USA*
 ***************************************************************************/

/* $Id: helicontrol.h 272 2008-12-17 22:50:20Z holgerf $ */

/** @file
 * API documentation of libhelicontrol, PLEASE READ ME!
 */

#ifndef HELICONTROL_H
#define HELICONTROL_H

#include <usb.h>
#include "helicontrol_firmware.h"

#ifdef __cplusplus
extern "C"
{
#endif


/** @name Picooz */
/*@{*/
/** use HC_PICOOZ_LEFT*step for yaw and just HC_PICOOZ_LEFT for trim */
#define HC_PICOOZ_LEFT                 -1
/** use HC_PICOOZ_RIGHT*step for yaw and just HC_PICOOZ_RIGHT for trim */
#define HC_PICOOZ_RIGHT                 1
/** maximum throttle value (helicontrol firmware will accept 15 as valid input, but actually send 14, for consistency with other models) */
#define HC_PICOOZ_MAX_THROTTLE         14
/** maximum absolute yaw value */
#define HC_PICOOZ_MAX_YAW               3
/** channel A */
#define HC_PICOOZ_CHANNEL_A             0
/** channel B */
#define HC_PICOOZ_CHANNEL_B             1
/** channel C */
#define HC_PICOOZ_CHANNEL_C             2
/*@}*/

/** @name Challenger */
/*@{*/
/** use HC_CHALLENGER_LEFT*step for yaw and just HC_CHALLENGER_LEFT for trim */
#define HC_CHALLENGER_LEFT             -1
/** use HC_CHALLENGER_RIGHT*step for yaw and just HC_CHALLENGER_RIGHT for trim */
#define HC_CHALLENGER_RIGHT             1
/** maximum throttle value */
#define HC_CHALLENGER_MAX_THROTTLE     15
/** maximum absolute yaw value */
#define HC_CHALLENGER_MAX_YAW           3
/** channel A */
#define HC_CHALLENGER_CHANNEL_A         1
/** channel B */
#define HC_CHALLENGER_CHANNEL_B         2
/*@}*/

/** @name Uranus */
/*@{*/
/** use HC_URANUS_LEFT*step for yaw and just HC_URANUS_LEFT for trim */
#define HC_URANUS_LEFT                 -1
/** use HC_URANUS_RIGHT*step for yaw and just HC_URANUS_RIGHT for trim */
#define HC_URANUS_RIGHT                 1
/** maximum throttle value */
#define HC_URANUS_MAX_THROTTLE         15
/** maximum absolute yaw value */
#define HC_URANUS_MAX_YAW               3
/** channel A */
#define HC_URANUS_CHANNEL_A             1
/** channel B */
#define HC_URANUS_CHANNEL_B             2
/** send normal pitch */
#define HC_URANUS_PITCH_NONE            0
/** send slow pitch boost */
#define HC_URANUS_PITCH_SLOW            1
/** send fast pitch boost */
#define HC_URANUS_PITCH_FAST            3
/** send continous pitch boost */
#define HC_URANUS_PITCH_CONT            2
/*@}*/

/** @name TandemZ */
/*@{*/
/** use HC_TANDEMZ_LEFT*step */
#define HC_TANDEMZ_LEFT                 1
/** use HC_TANDEMZ_RIGHT*step */
#define HC_TANDEMZ_RIGHT               -1
/** maximum throttle value */
#define HC_TANDEMZ_MAX_THROTTLE        15
/** maximum pitch trim value */
#define HC_TANDEMZ_MAX_TRIM            31
/** maximum pitch value */
#define HC_TANDEMZ_MAX_PITCH            5
/** maximum absolute yaw value */
#define HC_TANDEMZ_MAX_YAW             30
/** channel A */
#define HC_TANDEMZ_CHANNEL_A            0
/** channel B */
#define HC_TANDEMZ_CHANNEL_B            1
/** channel C */
#define HC_TANDEMZ_CHANNEL_C            2
/** center position for pitch (0..1 are backwards, 3..5 forwards) */
#define HC_TANDEMZ_PITCH_CENTER         2
/*@}*/

/** @name Flying Saucer */
/*@{*/
/** use HC_SAUCER_LEFT*step for yaw or roll */
#define HC_SAUCER_LEFT                 -1
/** use HC_SAUCER_RIGHT*step for yaw or roll */
#define HC_SAUCER_RIGHT                 1
/** use HC_SAUCER_FORWARD*step for pitch */
#define HC_SAUCER_FORWARD              -1
/** use HC_SAUCER_BACKWARD*step for pitch */
#define HC_SAUCER_BACKWARD              1
/** maximum throttle value */
#define HC_SAUCER_MAX_THROTTLE         15
/** maximum absolute pitch value */
#define HC_SAUCER_MAX_PITCH             7
/** maximum absolute yaw value */
#define HC_SAUCER_MAX_YAW               7
/** maximum absolute roll value */
#define HC_SAUCER_MAX_ROLL              7
/*@}*/



/**
 * try to open helicontrol usb device.
 * If more than one is connected, returns handle to first device which
 * has not been opened.
 * @return ptr to usb device handle, if one is found. NULL on error.
 * @see HC_Close
 */
struct usb_dev_handle *HC_Init(void);

/**
 * read firmware revision from device
 *
 * @param handle handle of usb device (obtained by HC_Init())
 * @param fwstring pointer to memory where revision string will be stored to (null-terminated).
 * @param maxlen maximum length of string, including tailing null.
 * @return something < 0 on error.
 */
int HC_Read_Firmware_Revision(usb_dev_handle *handle, char *fwstring, size_t maxlen);

int HC_Read_Firmware_Build_Date(usb_dev_handle *handle, char *datestring, size_t maxlen);

void HC_Read_Lib_Build_Date(char *datestring, size_t maxlen);

/**
 * switch continious sending. Note that all helis will stop after 300ms without command.
 *
 * @param handle handle of usb device (obtained by HC_Init())yy
 * @param repeat 0/1 enable or disable automatic repetition for the last command
 * @return number of bytes written/read or < 0 on error. See libusb/usb_control_msg() documentation.
 * @see HC_CMD_Interval
 */
int HC_CMD_Repeat(usb_dev_handle *handle, char repeat);

/**
 * specifies rate at which commands are sent.
 * @see HC_CMD_Repeat() which is used to enable continious sending.
 *
 * @param handle handle of usb device (obtained by HC_Init())
 * @param interval_ms interval in ms between end of a frame and start of the next one. If 0 the interval will be reset to its default value (200ms)
 * @return number of bytes written/read or < 0 on error. See libusb/usb_control_msg() documentation.
 * @see HC_CMD_Repeat
 */
int HC_CMD_Interval(usb_dev_handle *handle, int interval_ms);

/**
 * stop sending. Note that it is better to set throttle to 0 for an emergency stop of the heli. Otherwise
 * stop will occur approx. 300ms after calling HC_CMD_Stop().
 *
 * @param handle handle of usb device (obtained by HC_Init())
 * @return number of bytes written/read or < 0 on error. See libusb/usb_control_msg() documentation.
 */
int HC_CMD_Stop(usb_dev_handle *handle);

/**
 * PicooZ
 *
 * @param handle handle of usb device (obtained by HC_Init())
 * @param channel channel, 0..2 for A..C, see down side of heli
 * @param throttle throttle, 0..15
 * @param trim yaw trim, 1 = right, 0 = normal, -1 = left (trim value is stored in helicopter and only increased, when the trim bits are different from the previous frame)
 * @param yaw yaw, 1..3 left, 0 center, -1..-3 right
 * @return number of bytes written/read or < 0 on error. See libusb/usb_control_msg() documentation.
 */
int HC_Send_Picooz(usb_dev_handle *handle, char channel, char throttle, char yaw, char trim);

/**
 * Challanger
 *
 * @param handle handle of usb device (obtained by HC_Init())
 * @param channel channel, 0..2 for A..C, see down side of heli
 * @param throttle throttle, 0..15
 * @param trim yaw trim, 1 = right, 0 = normal, -1 = left (trim value is stored in helicopter and only increased, when the trim bits are different from the previous frame)
 * @param yaw yaw, 1..3 left, 0 center, -1..-3 right
 * @param fire 0/1, triggers "fire" LED in helicopter when set to 1
 * @return
 */
int HC_Send_Challenger(usb_dev_handle *handle, char channel, char throttle, char yaw, char trim, char fire);

/**
 * Uranus
 *
 * @param handle handle of usb device (obtained by HC_Init())
 * @param channel channel, 0..2 for A..C, see down side of heli
 * @param throttle throttle, 0..15
 * @param trim yaw trim, 1 = right, 0 = normal, -1 = left (trim value is stored in helicopter and only increased, when the trim bits are different from the previous frame)
 * @param yaw yaw, 1..3 left, 0 center, -1..-3 right
 * @param pitch 0..2, 0 = neutral/hover, 1 = slow boost, 2 = fast boost, 3 = continuous boost
 * @param light 0/1, turns on light when 1
 * @return number of bytes written/read or < 0 on error. See libusb/usb_control_msg() documentation.
 */
int HC_Send_Uranus(usb_dev_handle *handle, char channel, char throttle, char yaw, char pitch, char trim, char light);

/**
 * TandemZ
 *
 * @param handle handle of usb device (obtained by HC_Init())
 * @param channel channel, 0..2 for A..C, see down side of heli
 * @param light 0/1, turns on light when 1
 * @param throttle throttle, 0..15
 * @param pitch 0..1 = backwards, 2 = neutral/hover, 3..5 = forward
 * @param pitch_trim 0..31, 0 = max. forward, 31 = max. backwards
 * @param yaw 0..31, 0 = full left, 31 = full forward, center depends on helicopter (the original rc adds yaw trim to this value), steps 3-5, 19-22 and 27-29 are not valid, the helicontrol firmware will map those values to the nearest valid step
 * @return number of bytes written/read or < 0 on error. See libusb/usb_control_msg() documentation.
 */
int HC_Send_Tandemz(usb_dev_handle *handle, char channel, char throttle, char yaw, char pitch, char pitch_trim, char light);

/**
 * i-Bird
 * checksum algorithm is undetermined, but helicontrol firmware supports bit encoding for iBird, so checksum could be calculated/brute forced by host application
 *
 * @param handle handle of usb device (obtained by HC_Init())
 * @param channel channel, 2 bits
 * @param throttle throttle, 0..15
 * @param yaw yaw, 3 bits
 * @param unknown 5 bits, unknown purpose
 * @param checksum 5 bit checksum, algorithm undetermined
 * @return number of bytes written/read or < 0 on error. See libusb/usb_control_msg() documentation.
 */
int HC_Send_Ibird(usb_dev_handle *handle, char channel, char throttle, char yaw, char unknown, char checksum);

/**
 * saucer
 *
 * @param handle handle of usb device (obtained by HC_Init)
 * @param roll roll, -7..7, negative = left, positive = right
 * @param pitch pitch, -7..7, negative = forward, positive = backward
 * @param throttle throttle, 0..15
 * @param yaw yaw, -7..7, negative = left, positive = right
 * @return number of bytes written/read or < 0 on error. See libusb/usb_control_msg() documentation.
 */
int HC_Send_Saucer(usb_dev_handle *handle, char throttle, char yaw, char pitch, char roll);

/**
 * Close device. Please call me on exit.
 *
 * @param handle handle of usb device (obtained by HC_Init())
 * @see HC_Init()
 */
void HC_Close(usb_dev_handle *handle);

#ifdef __cplusplus
}
#endif

#endif /* HELICONTROL_H */
